from mlgame.core.communication import TransitionCommManager
from mlgame.core.exceptions import ErrorEnum, GameError, MLProcessError
from mlgame.core.model import MLGameDataType
from mlgame.utils.logger import logger


import websockets


import asyncio


class WebSocketExecutor():
    def __init__(self, ws_uri, ws_comm: TransitionCommManager):
        # super().__init__(name="ws")
        logger.info("             ws_init ")
        self._proc_name = f"websocket({ws_uri}"
        self._ws_uri = ws_uri
        self._comm_manager = ws_comm
        self._recv_data_func = self._comm_manager.recv_from_game

    async def ws_start(self):
        async with websockets.connect(self._ws_uri, ping_interval=None) as websocket:
            logger.info("             ws_start")
            count = 0
            is_ready_to_end = False
            while 1:
                
                data = self._recv_data_func()
                # logger.debug()
                if data.type==MLGameDataType.END:
                    logger.debug("ws received from game:", data)
                    break
                elif isinstance(data, GameError):
                    logger.debug("ws received :", data)
                    await websocket.send(data.data())
                    # exit container
                    if data.error_type in [ErrorEnum.COMMAND_ERROR, ErrorEnum.GAME_EXEC_ERROR]:
                        await websocket.send(
                            {"type": "system_message", "data": {
                                "message": f"error in {data.error_type}"}}
                        )
                        break
                        # os.system("pgrep -f 'tail -f /dev/null' | xargs kill")
                elif isinstance(data, MLProcessError):

                    logger.debug("ws received :", data)
                    # await websocket.send(data.data())
                    # exit container
                    # if data.error_type in [ErrorEnum.COMMAND_ERROR, ErrorEnum.GAME_EXEC_ERROR]:
                    await websocket.send(
                        {"type": "system_message", "data": {
                            "message": f"error in {data.message}"}}
                    )
                    break
                elif data.type == MLGameDataType.GAME_RESULT:
                    # raise a flag to recv data
                    # TODO replace with core.model
                    is_ready_to_end = True
                    logger.debug("GAME_RESULT :", data)
                    await websocket.send(data.model_dump_json())
                else:
                    # print(data)
                    logger.debug("ws send :", data)
                    if data.type == MLGameDataType.GAME_INFO:
                        await websocket.send(data.model_dump_json())
                    elif data.type == MLGameDataType.GAME_PROGRESS:
                        data.data={'frame':data.data.frame}
                        await websocket.send(data.model_dump_json())
                    elif data.type == MLGameDataType.SYSTEM_MSG:
                        await websocket.send(data.model_dump_json())
                    # count += 1
                    pass
                    # print(f'Send to ws : {count}:{data.keys()}')
                    #
                # make sure webservice got game result then mlgame is able to close websocket
                
                    
            while is_ready_to_end:
                # wait for game_result
                ws_recv_data = await websocket.recv()
                logger.info(f"ws received from django: {ws_recv_data}")
                if ws_recv_data == "game_result":
                    logger.info(f"< {ws_recv_data}")
                    await websocket.close()
                    is_ready_to_end = False
                    logger.info("close ws ")

                    # else:
                    #     await websocket.send(json.dumps({}))
            # time.sleep(1)
            # await websocket.close()

    def run(self):
        self._comm_manager.start_recv_obj_thread()
        try:
            asyncio.get_event_loop().run_until_complete(self.ws_start())
        except Exception as e:
            # exception = TransitionProcessError(self._proc_name, traceback.format_exc())
            self._comm_manager.send_exception(
                f"exception on {self._proc_name}")
            # catch connection error
            logger.exception(e)
        finally:
            logger.info("end ws ")


from os import path

import pygame
from mlgame.game.paia_game import PaiaGame, GameStatus
from mlgame.view.view_model import Scene

from .TeamBattleMode import TeamBattleMode
from .game_module.fuctions import get_sprites_progress_data

from .GenerateMap import MapGenerator

GAME_DIR = path.dirname(__file__)
MAP_DIR = path.join(GAME_DIR, "..", "asset", "maps")
SOUND_DIR = path.join(GAME_DIR, "..", "asset", "sound")
IMAGE_DIR = path.join(GAME_DIR, "..", "asset", "image")


class Game(PaiaGame):
    def __init__(self, user_num: int, green_team_num: int, blue_team_num: int, is_manual: str, frame_limit: int, sound: str):
        super().__init__(user_num)
        # init game
        self.green_team_num = green_team_num
        self.blue_team_num = blue_team_num
        self.is_paused = False
        self.is_debug = False
        self.is_sound = False
        self.is_manual = False
        if sound == "on":
            self.is_sound = True
        if is_manual:
            self.is_manual = True
        self.attachements = []
        self.frame_limit = frame_limit
        self.game_mode = self.set_game_mode()
        self.scene = Scene(width=self.game_mode.scene_width, height=self.game_mode.scene_height, color="#ffffff",
                           bias_y=50)

    def get_data_from_game_to_player(self) -> dict:
        to_players_data = self.game_mode.get_ai_data_to_player()
        return to_players_data

    def update(self, commands: dict):
        self.handle_event(commands)
        self.game_mode.debugging(self.is_debug)
        if not self.is_paused:
            self.frame_count += 1
            self.game_mode.update(commands)
            if not self.is_running():
                return "RESET"

    def reset(self):
        self.frame_count = 0
        self.game_mode.reset()
        self.rank()

    def get_scene_init_data(self) -> dict:
        """
        Get the scene and object information for drawing on the web
        """
        game_info = {'scene': self.scene.__dict__,
                     'assets': self.game_mode.get_init_image_data()}

        return game_info

    def get_scene_progress_data(self) -> dict:
        """
        Get the position of src objects for drawing on the web
        """
        scene_progress = {'background': [],
                          'object_list': [*self.game_mode.background, *self.get_obj_progress_data()],
                          'toggle_with_bias': [*self.game_mode.get_toggle_with_bias_data()],
                          'toggle': self.game_mode.get_toggle_progress_data(),
                          'foreground': [],
                          'user_info': [],
                          'game_sys_info': {}}

        return scene_progress

    def get_obj_progress_data(self):
        obj_list = []
        for sprites in self.game_mode.obj_list:
            obj_list.extend(get_sprites_progress_data(sprites))
        obj_list.extend(self.game_mode.obj_rect_list)
        return obj_list

    def get_game_result(self):
        """
        Get the src result for the web
        """
        self.rank()
        return {"frame_used": self.frame_count,
                "state": self.game_result_state,
                "attachment": self.attachements
                }

    def is_running(self):
        return self.game_mode.status == GameStatus.GAME_ALIVE

    def rank(self):
        self.game_result_state = self.game_mode.state
        self.attachements = self.game_mode.get_player_result()
        return self.attachements

    def handle_event(self, commands):
        if ["DEBUG"] in commands.values():
            self.is_debug = not self.is_debug
        if ["PAUSED"] in commands.values():
            self.is_paused = not self.is_paused

    def set_game_mode(self):
        sound_path = ""
        if self.is_sound:
            sound_path = SOUND_DIR
        map_width, map_height = MapGenerator(1, 1).getScreeenSize()
        play_rect_area = pygame.Rect(0, 0, map_width, map_height-100)
        game_mode = TeamBattleMode(self.green_team_num, self.blue_team_num, self.is_manual, self.frame_limit, sound_path, play_rect_area)
        return game_mode



import pygame
from .env import *

class Lane(pygame.sprite.Sprite):
    def __init__(self,y,distance):
        pygame.sprite.Sprite.__init__(self)
        self.rect = pygame.Rect(0, 0, lane_size[0], lane_size[1])
        self.rect.center = distance+150,y
        self.distance = distance

    def update(self, camera):
        if self.rect.right < 0:
            self.distance += 1150
        self.rect.centerx = self.distance-camera+450

    def get_asset_info(self):
        return {
            "type": "rect",
            "name": "lane",
            "color": WHITE,
            "x": self.rect.x,
            "y": self.rect.y,
            "width": lane_size[0],
            "height": lane_size[1],
            "angle": 0
        }

class Line(pygame.sprite.Sprite):
    def __init__(self, length):
        pygame.sprite.Sprite.__init__(self)
        self.rect = pygame.Rect(80, 114, 120, 450)
        self.type = "start_line"
        self.distance = 60
        self.end_distance = length

    def update(self, camera):
        if self.rect.right < 0:
            self.distance = self.end_distance
            self.type = "start_line"
        else:
            pass
        self.rect.left = self.distance - camera + 520

    def get_asset_info(self):
        return {
            "type": "image",
            "x": self.rect.x,
            "y": self.rect.y,
            "width": 120,
            "height": 450,
            "image_id": self.type,
            "angle": 0
        }


"""
The template of the main script of the machine learning process
"""
import pygame


class MLPlay:
    def __init__(self,ai_name, *args, **kwargs):
        """
        Constructor
        """
        self.ball_served = False

    def update(self, scene_info, keyboard=None, *args, **kwargs):
        """
        Generate the command according to the received `scene_info`.
        """
        # Make the caller to invoke `reset()` for the next round.
        if keyboard is None:
            keyboard = []
        if (scene_info["status"] == "GAME_OVER" or
                scene_info["status"] == "GAME_PASS"):
            return "RESET"

        if pygame.K_q in keyboard:
            command = "SERVE_TO_LEFT"
            self.ball_served = True
        elif pygame.K_e in keyboard:
            command = "SERVE_TO_RIGHT"
            self.ball_served = True
        elif pygame.K_LEFT in keyboard or pygame.K_a in keyboard:
            command = "MOVE_LEFT"
        elif pygame.K_RIGHT in keyboard or pygame.K_d in keyboard:
            command = "MOVE_RIGHT"
        else:
            command = "NONE"

        return command

    def reset(self):
        """
        Reset the status
        """
        self.ball_served = False



from os import path

import pygame
from mlgame.view.view_model import create_asset_init_data, create_image_view_data

from .env import IMAGE_DIR, WALL_LIVE


class Wall(pygame.sprite.Sprite):
    def __init__(self, construction, **kwargs):
        super().__init__()
        self.id = construction["_id"]
        self.no = 0
        self.rect = pygame.Rect(construction["_init_pos"], construction["_init_size"])
        self.angle = 0
        self.is_alive = True
        self.lives = WALL_LIVE

    def update(self, *args, **kwargs) -> None:
        if self.lives <= 0:
            self.kill()

    def collide_with_bullets(self):
        if self.lives > 0:
            self.lives -= 1

    def get_data_from_obj_to_game(self):
        info = {"id": f"wall_{self.lives}", "x": self.rect.x, "y": self.rect.y, "lives": self.lives}
        return info

    def get_obj_progress_data(self):
        return create_image_view_data(f"wall_{self.lives}", self.rect.x, self.rect.y
                                      , self.rect.width, self.rect.height, self.angle)

    def get_obj_init_data(self):
        image_init_data = []
        for i in range(1, self.lives+1):
            image_init_data.append(create_asset_init_data(f"wall_{i}", self.rect.width, self.rect.height,
                                                          path.join(IMAGE_DIR, f"wall_{min(i,3)}.png"),
                                                          f"https://raw.githubusercontent.com/Jesse-Jumbo/TankMan/main/asset/image/wall_{min(i,3)}.png"))
        return image_init_data



import pygame
from src import RacingCar

from src.env import FPS
from mlgame.view.view import PygameView
from mlgame.game.generic import quit_or_esc

if __name__ == '__main__':
    pygame.init()
    # game = RacingCar.RacingCar(user_num=2, game_mode="NORMAL", car_num=50, racetrack_length=10000, game_times=1, sound="off")
    game = RacingCar.RacingCar(user_num=1, game_type="RELIVE", car_num=20, racetrack_length=30000, rounds=1, sound="off")
    # game = RacingCar.RacingCar(2, "COIN", 30, 1000, 1, "off")
    scene_init_info_dict = game.get_scene_init_data()
    game_view = PygameView(scene_init_info_dict)
    frame_count = 0
    while game.isRunning() and not quit_or_esc():
        pygame.time.Clock().tick_busy_loop(FPS)
        game.update(game.get_keyboard_command())
        game_progress_data = game.get_scene_progress_data()
        game_view.draw(game_progress_data)
        frame_count += 1

    pygame.quit()



from mlgame.view.view_model import create_line_view_data, create_image_view_data
import pygame
from pygame import Rect, Surface
from pygame.math import Vector2
from pygame.sprite import Sprite
import random

from mlgame.game import physics
from mlgame.utils.enum import StringEnum, auto

class Brick(Sprite):
    def __init__(self, init_pos, *groups):
        super().__init__(*groups)
        self.rect = Rect(init_pos[0], init_pos[1], 25, 10)
        self.image = self._create_surface((244, 158, 66))   # Orange
        self.color = "#E09E42"


    def _create_surface(self, color):
        surface = Surface((self.rect.width, self.rect.height))
        surface.fill(color)
        
        # pygame.draw.line(surface, (0, 0, 0),
        #     (self.rect.width - 1, 0), (self.rect.width - 1, self.rect.height - 1))
        # pygame.draw.line(surface, (0, 0, 0),
        #     (0, self.rect.height - 1), (self.rect.width - 1, self.rect.height - 1))
        return surface

    @property
    def pos(self):
        return self.rect.topleft

    @property
    def get_line_data1(self):
        return create_line_view_data("brick_line1", self.rect.x + self.rect.width - 1, 
                                                    self.rect.y + 0, 
                                                    self.rect.x + self.rect.width - 1, 
                                                    self.rect.y + self.rect.height - 1, "#000000", 1)

    @property
    def get_line_data2(self):
        return create_line_view_data("brick_line2", self.rect.x + 0, 
                                                    self.rect.y + self.rect.height - 1 , 
                                                    self.rect.x +self.rect.width - 1, 
                                                    self.rect.y + self.rect.height - 1, "#000000", 1)

    @property
    def get_object_data(self):
        return create_image_view_data(
            "brick",
            self.rect.x,
            self.rect.y,
            self.rect.width,
            self.rect.height,
        )
        # return {"type": "rect",
        #         "name": "brick",
        #         "x": self.rect.x,
        #         "y": self.rect.y,
        #         "width": self.rect.width,
        #         "height": self.rect.height,
        #         "color": self.color}

class HardBrick(Brick):
    def __init__(self, init_pos, *groups):
        super().__init__(init_pos, *groups)

        self.reset()

    def reset(self):
        self.hp = 2
        # Override the origin color
        self.image = self._create_surface((209, 31, 31))    # Red
        self.color = "#D11F1F"

    def hit(self):
        """
        Decrease 1 HP and change the color of image and return the remaining HP

        @return The remaining HP
        """
        self.hp -= 1
        self.image = self._create_surface((244, 158, 66))   # Orange
        self.color = "#E09E42"

        return self.hp
    
    @property
    def get_object_data(self):
        return create_image_view_data(
            "hard_brick",
            self.rect.x,
            self.rect.y,
            self.rect.width,
            self.rect.height,
        )

class PlatformAction(StringEnum):
    SERVE_TO_LEFT = auto()
    SERVE_TO_RIGHT = auto()
    MOVE_LEFT = auto()
    MOVE_RIGHT = auto()
    NONE = auto()

SERVE_BALL_ACTIONS = (PlatformAction.SERVE_TO_LEFT, PlatformAction.SERVE_TO_RIGHT)

class Platform(Sprite):
    def __init__(self, init_pos, play_area_rect: Rect, *groups):
        super().__init__(*groups)

        self._play_area_rect = play_area_rect
        self._shift_speed = 5
        self._speed = [0, 0]
        self._init_pos = init_pos

        self.rect = Rect(init_pos[0], init_pos[1], 40, 10)
        self.image = self._create_surface()

    def _create_surface(self):
        surface = Surface((self.rect.width, self.rect.height))
        surface.fill((66, 226, 126)) # Green
        self.color = "#42E27E"
        return surface

    @property
    def pos(self):
        return self.rect.topleft

    def reset(self):
        self.rect.topleft = self._init_pos

    def move(self, move_action: PlatformAction):
        if (move_action == PlatformAction.MOVE_LEFT and
            self.rect.left > self._play_area_rect.left):
            self._speed[0] = -self._shift_speed
        elif (move_action == PlatformAction.MOVE_RIGHT and
              self.rect.right < self._play_area_rect.right):
            self._speed[0] = self._shift_speed
        else:
            self._speed[0] = 0

        self.rect.move_ip(*self._speed)

    @property
    def get_object_data(self):
        return create_image_view_data(
            "board",
            self.rect.x,
            self.rect.y,
            self.rect.width,
            self.rect.height,
        )

class Ball(Sprite):
    def __init__(self, init_pos, play_area_rect: Rect, *groups):
        super().__init__(*groups)

        self._play_area_rect = play_area_rect
        self._do_slide_ball = True
        self._init_pos = init_pos
        self._speed = [0, 0]

        self.hit_platform_times = 0

        self.rect = Rect(*self._init_pos, 10, 10)
        self.image = self._create_surface()

        # For additional collision checking
        self._last_pos = self.rect.copy()

        self.hit_brick_false = 0


    def _create_surface(self):
        surface = pygame.Surface((self.rect.width, self.rect.height))
        surface.fill((44, 185, 214)) # Blue
        self.color = "#2CB9D6"
        return surface

    @property
    def pos(self):
        return self.rect.topleft

    def reset(self):
        self.hit_platform_times = 0
        self.rect.topleft = self._init_pos
        self._speed = [0, 0]

    def stick_on_platform(self, platform_centerx):
        self.rect.centerx = platform_centerx

    def serve(self, platform_action: PlatformAction):
        if platform_action == PlatformAction.SERVE_TO_LEFT:
            self._speed = [-7, -7]
        elif platform_action == PlatformAction.SERVE_TO_RIGHT:
            self._speed = [7, -7]


    def move(self):
        self._last_pos.topleft = self.rect.topleft
        self.rect.move_ip(self._speed)

    def check_bouncing(self, platform: Platform):
        if (physics.collide_or_contact(self, platform) or
                self._platform_additional_check(platform)):
            self.hit_platform_times += 1
            self.hit_brick_false += 1

            rect_after_bounce, speed_after_bounce = physics.bounce_off(
                self.rect, self._speed, platform.rect, platform._speed)
            # Check slicing ball when the ball goes up after bouncing (not game over)
            if self._do_slide_ball and speed_after_bounce[1] < 0:
                speed_after_bounce[0] = self._slice_ball(self._speed[0], platform._speed[0])

            self.rect = rect_after_bounce
            self._speed = speed_after_bounce

        if physics.rect_break_or_contact_box(self.rect, self._play_area_rect):
            physics.bounce_in_box_ip(self.rect, self._speed, self._play_area_rect)

    def _platform_additional_check(self, platform: Platform):
        """
        The additional checking for the condition that the ball passes the corner of the platform
        """
        if self.rect.bottom > platform.rect.top:
            routine_a = (Vector2(self._last_pos.bottomleft), Vector2(self.rect.bottomleft))
            routine_b = (Vector2(self._last_pos.bottomright), Vector2(self.rect.bottomright))

            return (physics.rect_collideline(platform.rect, routine_a) or
                    physics.rect_collideline(platform.rect, routine_b))

        return False

    def _slice_ball(self, ball_speed_x, platform_speed_x):
        """
        Check if the platform slices the ball, and modify the ball speed.

        @return The new x speed of the ball after slicing
        """
        # If the platform doesn't move, bounce normally.
        if platform_speed_x == 0:
            return 7 if ball_speed_x > 0 else -7
        # If the platform moves at the same direction as the ball moving,
        # speed up the ball.
        elif ball_speed_x * platform_speed_x > 0:
            return 10 if ball_speed_x > 0 else -10
        # If the platform moves at the opposite direction against the ball moving,
        # hit the ball back.
        else:
            return -7 if ball_speed_x > 0 else 7

    def check_hit_brick(self, group_brick: pygame.sprite.RenderPlain) -> int:
        """
        Check if the ball hits bricks in the `group_brick`.
        The hit bricks will be removed from `group_brick`, but the alive hard brick will not.
        However, if the ball speed is high, the hard brick will be rdedmoved with only one hit.

        @param group_brick The sprite group containing bricks
        @return destroyed bricks and created bricks
        """
        hit_bricks = pygame.sprite.spritecollide(self, group_brick, 1,
                                                 physics.collide_or_contact)
        new_bricks = []

        num_of_destroyed_brick = len(hit_bricks)
        if hit_bricks:
            self.hit_brick_false = 0


        if num_of_destroyed_brick > 0:
            # XXX: Bad multiple collision bouncing handling
            if (num_of_destroyed_brick == 2 and
                    (hit_bricks[0].rect.y == hit_bricks[1].rect.y or
                     hit_bricks[0].rect.x == hit_bricks[1].rect.x)):
                combined_rect = hit_bricks[0].rect.union(hit_bricks[1].rect)
                physics.bounce_off_ip(self.rect, self._speed, combined_rect, (0, 0))
            else:
                physics.bounce_off_ip(self.rect, self._speed, hit_bricks[0].rect, (0, 0))

            if abs(self._speed[0]) == 7:
                for brick in hit_bricks:
                    if isinstance(brick, HardBrick) and brick.hit():
                        new_bricks.append(Brick(brick.pos,group_brick))
                        num_of_destroyed_brick -= 1

        return hit_bricks, new_bricks

    @property
    def get_object_data(self):
        return create_image_view_data(
            "ball",
            self.rect.x,
            self.rect.y,
            self.rect.width,
            self.rect.height,
        )
        # return {"type": "rect",
        #         "name": "ball",
        #         "x": self.rect.x,
        #         "y": self.rect.y,
        #         "width": self.rect.width,
        #         "height": self.rect.height,
        #         "color": self.color}



import importlib
import os
from argparse import ArgumentParser
from os import path

from pydantic import ValidationError

from mlgame.argument.tool import get_data_from_json_file
from mlgame.argument.model import UserNumConfig
from mlgame.core.exceptions import GameConfigError


def create_game_params_parser(parser_config: dict):
    """
    Generate `argparse.ArgumentParser` from `parser_config`

    @param parser_config A dictionary carries parameters for creating `ArgumentParser`.
           The key "()" specifies parameters for constructor of `ArgumentParser`,
           its value is a dictionary of which the key is the name of parameter and
           the value is the value to be passed to that parameter.
           The remaining keys of `parser_config` specifies arguments to be added to the parser,
           which `ArgumentParser.add_argument() is invoked`. The key is the name
           of the argument, and the value is similar to the "()"
           but for the `add_argument()`. Note that the name of the key is used as the name
           of the argument, but if "name_or_flags" is specified in the dictionary of it,
           it will be passed to the `add_argument()` instead. The value of "name_or_flags"
           must be a tuple.
           An example of `parser_config`:
           ```
            {
                "()": {
                    "usage": "game <difficulty> <level>"
                },
                "difficulty": {
                    "choices": ["EASY", "NORMAL"],
                    "metavar": "difficulty",
                    "help": "Specify the game style. Choices: %(choices)s"
                },
                "level": {
                    "type": int,
                    "help": "Specify the level map"
                },
            }
           ```
    """
    if parser_config.get("()"):
        parser = ArgumentParser(**parser_config["()"])
    else:
        parser = ArgumentParser()

    for arg_name in parser_config.keys():
        if arg_name != "()":
            arg_config = parser_config[arg_name].copy()

            name_or_flag = arg_config.pop("name_or_flags", None)
            if not name_or_flag:
                name_or_flag = (arg_name,)

            parser.add_argument(*name_or_flag, **arg_config)

    return parser

def comma_separated_list(value):
    return value.split(',')
def parse_game_config_data(game_config_data: dict):
    """
    parse game parameter and generate data for argument parser
    """
    # TODO to optimize
    result = {}
    params = game_config_data["game_params"]
    game_usage = "%(prog)s "
    for param in params:
        obj = {
            "metavar": param["verbose"],
            "help": param["help"]

        }
        if param["type"] == "int":
            obj["type"] = int
        elif param["type"] == "str":
            obj["type"] = str
        elif param["type"] == "list":
            obj["type"] = comma_separated_list
        elif param["type"] == "path":
            obj["type"] = os.path.abspath

        if "default" in param:
            obj["nargs"] = "?"
            obj["default"] = param["default"]
            game_usage += "[" + param["name"] + "] "
        else:
            game_usage += "<" + param["name"] + "> "

        if "choices" in param:
            choices = []
            for choice in param["choices"]:
                if type(choice) == dict:
                    choices.append(choice["value"])
                else:
                    choices.append(choice)
            obj["choices"] = choices
        if "min" in param and "max" in param:
            obj["choices"] = range(param["min"], param["max"] + 1)
        """
        ex -t --time_to_play
        """
        if "flag" in param:
            obj["name_or_flags"] = (f'-{param["flag"]}', f'--{param["name"]}')
        else:
            obj["name_or_flags"] = (f'--{param["name"]}',)

        result[param["name"]] = obj
    result["()"] = {
        "prog": game_config_data["game_name"],
        "game_usage": game_usage
    }
    return result


class GameConfig:
    """
    The data class storing the game defined config
    Included game_config.json game_cls
    """

    def __init__(self, game_folder: str):
        """
        Parse the game defined config and generate a `GameConfig` instance
        """
        game_config = self._load_game_config(game_folder)

        config_data = get_data_from_json_file(path.join(game_folder, "game_config.json"))
        try:
            self.game_version = config_data["version"]
            self.game_name = config_data["game_name"]
            self.user_num_config = UserNumConfig(**config_data["user_num"])
            self._process_game_param_dict(config_data)
            self.game_config_parser = create_game_params_parser(self._config_to_create_parser)
            self.game_cls = None
            self.logo = None
            logo_path = config_data.get("logo", [])
            for logo in logo_path:
                if not str(logo).startswith("http"):
                    self.logo = path.join(game_folder, logo)
                    break
            game_setup = getattr(game_config, "GAME_SETUP")
            self.game_cls = game_setup["game"]
        except AttributeError:
            raise GameConfigError("Missing variable 'GAME_SETUP' in the config.py")
        except KeyError:
            raise GameConfigError(f"game_config.json in {game_folder} "
                                  f"should contains 'user_num', 'version', 'game_params' ")
        except ValidationError:
            raise GameConfigError("`user_num` in game_config.json should contains 'min' and 'max' "
                                  "and user_num['min'] < user_num['max']")
        # self._process_game_setup_dict()

    def parse_game_params(self, game_params) -> dict:
        return self.game_config_parser.parse_args(game_params).__dict__

    def _load_game_config(self, game_folder):
        """
        Load the game config
        """
        try:
            # game_config = importlib.import_module(f"{game_folder}.config")

            spec = importlib.util.spec_from_file_location("config", path.join(game_folder, "config.py").__str__())
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            game_config = module
        except ModuleNotFoundError as e:
            # print which module is not found or installed at which game_folder
            failed_module_name = e.__str__().split("'")[1]
            msg = f"Module '{failed_module_name}' is not found in game process"
            # TODO error提供資訊不足
            raise GameConfigError(msg)
        else:
            return game_config

    def _process_game_param_dict(self, config_data):
        """
        Convert some fields in `GAME_PARAMS`
        """
        self._config_to_create_parser = parse_game_config_data(config_data)

        # Append the prefix of MLGame.py usage to the `game_usage`
        # and set it to the `usage`
        if self._config_to_create_parser.get("()") and self._config_to_create_parser["()"].get("game_usage"):
            game_usage = str(self._config_to_create_parser["()"].pop("game_usage"))
            self._config_to_create_parser["()"]["usage"] = "python MLGame.py [options] " + game_usage

        # If the game not specify "--version" flag,
        # try to convert `GAME_VERSION` to a flag
        if not self._config_to_create_parser.get("--version"):
            self._config_to_create_parser["--version"] = {
                "action": "version",
                "version": self.game_version
            }



"""
The subprocess for running the client script and handling the I/O of it
"""
import json

from subprocess import PIPE, Popen
from threading import Thread, Event
from queue import Queue

from .exceptions import MLClientExecutionError

class Client(Popen):
    """
    The subprocess for executing and communication with non-python client
    """

    def __init__(self, execution_cmd: list):
        """
        Constructor
        """
        super().__init__(execution_cmd, bufsize = 1,
            stdin = PIPE, stdout = PIPE, stderr = PIPE,
            universal_newlines = True)

        # Thread for reading message
        self._command_obj_queue = Queue()
        self._is_program_exited = Event()
        self._read_stdout_thread = Thread(target = self._read_stdout,
            name = "read_stdout")

        self._read_stdout_thread.start()

    def send_to_client(self, header, dict_payload):
        """
        Send the dictionary object to the client
        The object will be converted into a string:
        "<header> <JSON_representation_of_dict_object>".

        @param header The header to be added at the begin of the message
        @param dict_payload A dictionay object
        """
        if not self._is_program_exited.is_set():
            self.stdin.write(header + " " + json.dumps(dict_payload) + "\n")
            self.stdin.flush()

    def recv_from_client(self):
        """
        Receive the command from the subprocess
        It will wait until there has command arrived.
        """
        return self._command_obj_queue.get()

    def _read_stdout(self):
        """
        Read the message from stdout of the client.
        If the message contains "__command__" header, it will be pushed to
        the command object queue. Otherwise, it will be printed out.
        """
        while True:
            message = self.stdout.readline()
            if "__command__" in message:
                # Remove header and the trailing newline
                message = message[12:-1]
                if message == "READY" or message == "RESET":
                    self._command_obj_queue.put(message)
                else:
                    self._command_obj_queue.put(json.loads(message))
            elif len(message):
                print(message, flush = True)
            else:   # zero-length read, program is exited
                break

        self._is_program_exited.set()
        self.wait()

        if self.returncode != 0:
            message = " Get output from stderr:\n" + self._read_stderr()
        else:
            message = ""

        self._command_obj_queue.put(MLClientExecutionError(
            "The user program is exited with returncode {}.{}"
            .format(self.returncode, message)))

    def _read_stderr(self):
        """
        Read the message from stderr of the client.
        """
        message = self.stderr.readlines()
        return "".join(message)
